@@ -1,26 +1,29 @@
-;;; @(#) ada-prj.el --- Easy editing of project files for the ada-mode
+;;; ada-prj.el --- easy editing of project files for the ada-mode
 
-;; Copyright (C) 1998-1999 Free Software Foundation, Inc.
+;; Copyright (C) 1998, 99, 2000, 2001 Free Software Foundation, Inc.
 
 ;; Author: Emmanuel Briot <briot@gnat.com>
-;; Ada Core Technologies's version:   $Revision: 1.48 $
+;; Ada Core Technologies's version:   $Revision: 1.53 $
 ;; Keywords: languages, ada, project file
 
-;; This file is not part of GNU Emacs.
+;; This file is part of GNU Emacs.
 
-;; This program is free software; you can redistribute it and/or modify
+;; GNU Emacs is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
 ;; the Free Software Foundation; either version 2, or (at your option)
 ;; any later version.
 
-;; This program is distributed in the hope that it will be useful,
+;; GNU Emacs is distributed in the hope that it will be useful,
 ;; but WITHOUT ANY WARRANTY; without even the implied warranty of
 ;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with GNU Emacs; see the file COPYING.  If not, write to
-;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
+;; along with GNU Emacs; see the file COPYING.  If not, write to the
+;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
+;; Boston, MA 02111-1307, USA.
+
+;;; Commentary:
 
 ;;; This package provides a set of functions to easily edit the project
 ;;; files used by the ada-mode.
@@ -30,7 +33,7 @@
 ;;; Internally, a project file is represented as a property list, with each
 ;;; field of the project file matching one property of the list.
 
-;; Code:
+;;; Code:
 
 
 ;; ----- Requirements -----------------------------------------------------
@@ -50,6 +53,9 @@
 (defvar ada-prj-ada-buffer nil
   "Indicates what Ada source file was being edited.")
 
+(defvar ada-old-cross-prefix nil
+  "The cross-prefix associated with the currently loaded runtime library.")
+
 
 ;; ----- Functions --------------------------------------------------------
 
@@ -139,6 +145,7 @@ If the current value of FIELD is the default value, returns an empty string."
 	  (ada-prj-save-specific-option 'bind_opt)
 	  (ada-prj-save-specific-option 'link_opt)
 	  (ada-prj-save-specific-option 'gnatmake_opt)
+	  (ada-prj-save-specific-option 'gnatfind_opt)
 	  (ada-prj-save-specific-option 'cross_prefix)
 	  (ada-prj-save-specific-option 'remote_machine)
 	  (ada-prj-save-specific-option 'debug_cmd)
@@ -155,9 +162,6 @@ If the current value of FIELD is the default value, returns an empty string."
 			    (plist-get ada-prj-current-values 'comp_cmd)) "\n"
 	  (ada-prj-set-list "run_cmd"
 			    (plist-get ada-prj-current-values 'run_cmd)) "\n"
-	  (ada-prj-set-list "casing"
-			    (plist-get ada-prj-current-values 'casing)
-			    t) "\n"
 	  (ada-prj-set-list "src_dir"
 			    (plist-get ada-prj-current-values 'src_dir)
 			    t) "\n"
@@ -251,6 +255,14 @@ The current buffer must be the project editing buffer."
   (let ((inhibit-read-only t))
     (erase-buffer))
 
+  ;;  Widget support in Emacs 21 requires that we clear the buffer first
+  (if (and (not (boundp 'running-xemacs)) (>= emacs-major-version 21))
+      (progn
+	(setq widget-field-new  nil
+	      widget-field-list nil)
+	(mapcar (lambda (x) (delete-overlay x)) (car (overlay-lists)))
+	(mapcar (lambda (x) (delete-overlay x)) (cdr (overlay-lists)))))
+  
   ;;  Display the tabs
   
   (widget-insert "\n               Project and Editor configuration.\n
@@ -326,6 +338,14 @@ To use JGNAT, enter 'j'.")
    ;;  Second page (Paths)
    ;;
    ((= tab-num 2)
+    (if (not (equal (plist-get ada-prj-current-values 'cross_prefix)
+		    ada-old-cross-prefix))
+	(progn
+	  (setq ada-old-cross-prefix
+		(plist-get ada-prj-current-values 'cross_prefix))
+	  (ada-initialize-runtime-library ada-old-cross-prefix)))
+
+    
     (widget-insert "/_____________\\/              \\/______________\\/______________\\/______________\\\n")
     (ada-prj-field 'src_dir  "Source directories"
 "Enter the list of directories where your Ada
@@ -376,6 +396,11 @@ command and are passed to the linker")
     (ada-prj-field 'gnatmake_opt "Switches for gnatmake"
 "These switches are used in the default gnatmake
 command.")		   
+    (ada-prj-field 'gnatfind_opt "Switches for gnatfind"
+"The command gnatfind is run every time the Ada/Goto/List_References menu.
+You should for instance add -a if you are working in an environment
+where most ALI files are write-protected, since otherwise they get
+ignored by gnatfind and you don't see the references within.")
     )
 
    ;;
@@ -389,11 +414,14 @@ where name is the name that appears after the Help buttons in this buffer. As
 a special case, ${current} is replaced with the name of the file currently
 edited, with directory name but no extension, whereas ${full_current} is
 replaced with the name of the current file with directory name and
-extension.\n\n")
+extension.\n")
     (widget-insert
 "The environment variables ADA_INCLUDE_PATH and ADA_OBJECTS_PATH are set to
 ${src_dir} and ${obj_dir} before running the compilation commands, so that you
-don't need to specify the -aI and -aO switches on the command line\n\n")
+don't need to specify the -aI and -aO switches on the command line\n")
+    (widget-insert
+"You can reference any environment variable using the same ${...} syntax as
+above, and put the name of the variable between the quotes.\n\n")
     (ada-prj-field 'check_cmd
       "Check syntax of a single file (menu Ada->Check File)"
 "This command is run to check the syntax and semantics of a file.
@@ -488,6 +516,9 @@ If FILENAME is given, edit that file."
     (make-local-variable 'widget-keymap)
     (define-key widget-keymap "\C-x\C-s" 'ada-prj-save)
 
+    (set (make-local-variable 'ada-old-cross-prefix)
+	 (ada-xref-get-project-field 'cross-prefix))
+
     (ada-prj-display-page 1)
   ))
 
@@ -531,7 +562,7 @@ change in ada-prj-current-values so that selecting another page and coming
 back keeps the new value."
   (set 'ada-prj-current-values
        (plist-put ada-prj-current-values
-		  (widget-get widget 'prj-field)
+		  (widget-get widget ':prj-field)
 		  (widget-value widget))))
 
 (defun ada-prj-display-help (widget widget-modified event)
@@ -552,15 +583,17 @@ this function can be used as :notify for the widget."
       )))
 
 (defun ada-prj-show-value (widget widget-modified event)
-  (let ((value (plist-get ada-prj-current-values
-			  (widget-get widget 'prj-field)))
-	(inhibit-read-only t))
+  (let* ((field (widget-get widget ':prj-field))
+	 (value (plist-get ada-prj-current-values field))
+	 (inhibit-read-only t)
+	 w)
 
     ;;  If the other widget is already visible, delete it
     (if (widget-get widget 'prj-other-widget)
 	(progn
 	  (widget-delete (widget-get widget 'prj-other-widget))
 	  (widget-put widget 'prj-other-widget nil)
+	  (widget-put widget ':prj-field field)
 	  (widget-default-value-set widget "Show Value")
 	  )
 
@@ -569,14 +602,15 @@ this function can be used as :notify for the widget."
 	(mouse-set-point event)
 	(forward-line 1)
 	(beginning-of-line)
-	(widget-put widget 'prj-other-widget
-		    (widget-create 'editable-list
-				   :entry-format "%i%d %v"
-				   :notify 'ada-prj-field-modified
-				   :help-echo (widget-get widget 'prj-help)
-				   :value value
-				   (list 'editable-field
-					 :keymap widget-keymap)))
+	(setq w (widget-create 'editable-list
+			       :entry-format "%i%d %v"
+			       :notify 'ada-prj-field-modified
+			       :help-echo (widget-get widget 'prj-help)
+			       :value value
+			       (list 'editable-field :keymap widget-keymap)))
+	(widget-put widget 'prj-other-widget w)
+	(widget-put w ':prj-field field)
+	(widget-put widget ':prj-field field)
 	(widget-default-value-set widget "Hide Value")
 	)
       )
@@ -622,6 +656,7 @@ AFTER-TEXT is inserted just after the widget."
 				     (list 'quote field)))
 			 "Load Recursive Directory")
 	  (widget-insert "\n           ${build_dir}\n")))
+
     (set 'widget
 	 (if is-list
 	     (if (< (length value) 15)
@@ -631,11 +666,11 @@ AFTER-TEXT is inserted just after the widget."
 				:help-echo help-text
 				:value value
 				(list 'editable-field :keymap widget-keymap))
+
 	       (let ((w (widget-create 'push-button
 				       :notify 'ada-prj-show-value
 				       "Show value")))
 		 (widget-insert "\n")
-		 (widget-put w 'prj-field field)
 		 (widget-put w 'prj-help  help-text)
 		 (widget-put w 'prj-other-widget nil)
 		 w)
@@ -646,7 +681,7 @@ AFTER-TEXT is inserted just after the widget."
 			  :help-echo help-text
 			  :keymap widget-keymap
 			  value)))
-    (widget-put widget 'prj-field field)
+    (widget-put widget ':prj-field field)
     (if after-text
 	(widget-insert after-text))
     (widget-insert "\n")
@@ -658,4 +693,5 @@ AFTER-TEXT is inserted just after the widget."
 (ada-prj-add-keymap)
 
 (provide 'ada-prj)
-;;; package ada-prj.el ends here
+
+;;; ada-prj.el ends here
