@@ -1,12 +1,14 @@
-;;; ada-stmt.el - An extension to Ada mode for inserting statement templates.
+;;; ada-stmt.el --- an extension to Ada mode for inserting statement templates
 
 ;; Copyright(C) 1987, 93, 94, 96, 97, 98, 99, 2000
 ;;   Free Software Foundation, Inc.
 
-;; Ada Core Technologies's version:   $Revision: 1.18 $ 
+;; Ada Core Technologies's version:   $Revision: 1.21 $ 
+
+;; This file is part of GNU Emacs.
 
 ;; Authors: Daniel Pfeiffer, Markus Heritsch, Rolf Ebert <ebert@waporo.muc.de>
-;; Maintainer: Rolf Ebert <ebert@waporo.muc.de>
+;; Maintainer: Emmanuel Briot <briot@gnat.com>
 ;; Keywords: languages, ada
 ;; Rolf Ebert's version: 2.26
 
@@ -108,14 +110,14 @@
 		 ["When" ada-when t])))
     (if ada-xemacs
 	(funcall (symbol-function 'add-submenu)
-		 '("Ada") (append (list "Statements"
+		 '("Ada") (append (list "Templates"
 					:included '(string= mode-name "Ada"))
 				  menu))
 
-      (define-key-after (lookup-key ada-mode-map [menu-bar Ada]) [Statements]
+      (define-key-after (lookup-key ada-mode-map [menu-bar Ada]) [Templates]
 	(list 'menu-item
-	      "Statements"
-	      (easy-menu-create-menu "Statements" menu)
+	      "Templates"
+	      (easy-menu-create-menu "Templates" menu)
 	      :visible '(string= mode-name "Ada"))
 	t))))
 
@@ -127,7 +129,7 @@
   (save-excursion
     (let ((case-fold-search t))
       (if (re-search-backward ada-procedure-start-regexp nil t)
-	  (buffer-substring (match-beginning 2) (match-end 2))
+	  (buffer-substring (match-beginning 3) (match-end 3))
 	"NAME?"))))
 
 (defvar ada-template-map
@@ -566,8 +568,7 @@ Invoke right after `ada-function-spec' or `ada-procedure-spec'."
     (let ((aa-end (point)))
       (ada-adjust-case-region
        (progn (goto-char (symbol-value 'beg)) (forward-word -1) (point))
-       (goto-char aa-end))
-      )))
+       (goto-char aa-end)))))
 
 (defun ada-stmt-mode-hook ()
   (set (make-local-variable 'skeleton-further-elements)
@@ -576,7 +577,7 @@ Invoke right after `ada-function-spec' or `ada-procedure-spec'."
   (add-hook 'skeleton-end-hook
 	    'ada-adjust-case-skeleton nil t)
   (ada-stmt-add-to-ada-menu))
-       
+
 (add-hook 'ada-mode-hook 'ada-stmt-mode-hook)
 
 (provide 'ada-stmt)
