@@ -1,4 +1,4 @@
-;; @(#) ada-xref.el --for lookup and completion in Ada mode
+;;; ada-xref.el --- for lookup and completion in Ada mode
 
 ;; Copyright (C) 1994, 95, 96, 97, 98, 99, 2000, 2001
 ;;    Free Software Foundation, Inc.
@@ -7,24 +7,25 @@
 ;;      Rolf Ebert <ebert@inf.enst.fr>
 ;;      Emmanuel Briot <briot@gnat.com>
 ;; Maintainer: Emmanuel Briot <briot@gnat.com>
-;; Ada Core Technologies's version:   $Revision: 1.124 $
+;; Ada Core Technologies's version:   $Revision: 1.151 $
 ;; Keywords: languages ada xref
 
 ;; This file is part of GNU Emacs.
 
-;; This program is free software; you can redistribute it and/or modify
+;; GNU Emacs is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
 ;; the Free Software Foundation; either version 2, or (at your option)
 ;; any later version.
 
-;; This program is distributed in the hope that it will be useful,
+;; GNU Emacs is distributed in the hope that it will be useful,
 ;; but WITHOUT ANY WARRANTY; without even the implied warranty of
 ;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with GNU Emacs; see the file COPYING.  If not, write to
-;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
+;; along with GNU Emacs; see the file COPYING.  If not, write to the
+;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
+;; Boston, MA 02111-1307, USA.
 
 ;;; Commentary:
 ;;; This Package provides a set of functions to use the output of the
@@ -37,6 +38,8 @@
 
 ;;; You need Emacs >= 20.2 to run this package
 
+;;; Code:
+
 ;; ----- Requirements -----------------------------------------------------
 
 (require 'compile)
@@ -88,9 +91,10 @@ The command gnatfind is used every time you choose the menu
   :type 'string :group 'ada)
 
 (defcustom ada-prj-default-comp-cmd
-  "${cross_prefix}gcc -c ${comp_opt} ${full_current}"
+  "${cross_prefix}gcc -x ada -c ${comp_opt} ${full_current}"
   "*Default command to be used to compile a single file.
-This is the same syntax as in the project file."
+Emacs will add the filename at the end of this command. This is the same
+syntax as in the project file."
   :type 'string :group 'ada)
 
 (defcustom ada-prj-default-debugger "${cross_prefix}gdb"
@@ -105,10 +109,10 @@ This is the same syntax as in the project file."
 This is the same syntax as in the project file."
   :type 'string :group 'ada)
 
-(defcustom ada-prj-default-project-file nil
+(defcustom ada-prj-default-project-file ""
   "*Name of the project file to use for every Ada file.
-Nil means that a default project file will be provided by Emacs the first
-time such a file is needed."
+Emacs will not try to use the standard algorithm to find the project file if
+this string is not empty."
   :type '(file :must-match t) :group 'ada)
 
 (defcustom ada-gnatstub-opts "-q -I${src_dir}"
@@ -121,6 +125,20 @@ This has the same syntax as in the project file (with variable substitution)."
 Otherwise, ask the user for the name of the project file to use."
   :type 'boolean :group 'ada)
 
+(defconst is-windows (memq system-type (quote (windows-nt)))
+  "True if we are running on windows NT or windows 95.")
+
+(defcustom ada-tight-gvd-integration nil
+  "*If non-nil, a new Emacs frame will be swallowed in GVD when debugging.
+If GVD is not the debugger used, nothing happens.")
+
+(defcustom ada-xref-search-with-egrep t
+  "*If non-nil, use egrep to find the possible declarations for an entity.
+This alternate method is used when the exact location was not found in the
+information provided by GNAT. However, it might be expensive if you have a lot
+of sources, since it will search in all the files in your project."
+  :type 'boolean :group 'ada)
+
 ;; ------- Nothing to be modified by the user below this
 (defvar ada-last-prj-file ""
   "Name of the last project file entered by the user.")
@@ -131,9 +149,6 @@ Otherwise, ask the user for the name of the project file to use."
 (defconst ada-project-file-extension ".adp"
   "The extension used for project files.")
 
-(defconst is-windows (memq system-type (quote (windows-nt)))
-  "True if we are running on windows NT or windows 95.")
-
 (defvar ada-xref-runtime-library-specs-path '()
   "Directories where the specs for the standard library is found.
 This is used for cross-references.")
@@ -154,12 +169,11 @@ Used to go back to these positions.")
 using cmdproxy.exe as the shell, we need to use /d or the drive is never
 changed.")
 
-(defvar ada-command-separator
-  (if (string-match "cmdproxy.exe" shell-file-name) " && " "\n")
+(defvar ada-command-separator (if is-windows " && " "\n")
   "Separator to use when sending multiple commands to `compile' or
 `start-process'.
 cmdproxy.exe doesn't recognize multiple-line commands, so we have to use
-\"\n\" for now.")
+\"&&\" for now.")
 
 (defconst ada-xref-pos-ring-max 16
   "Number of positions kept in the list ada-xref-pos-ring.")
@@ -171,14 +185,19 @@ cmdproxy.exe doesn't recognize multiple-line commands, so we have to use
 (defvar ada-xref-project-files '()
   "Associative list of project files.
 It has the following format:
-((project_name . value) (project_name . value) ...)
+\((project_name . value) (project_name . value) ...)
 As always, the values of the project file are defined through properties.")
 
-(defun ada-initialize-runtime-library ()
-  "Initializes the variables for the runtime library location."
+(defun ada-quote-cmd (cmd)
+  "Duplicates all \\ characters in CMD so that it can be passed to `compile'"
+  (mapconcat 'identity (split-string cmd "\\\\") "\\\\"))
+
+(defun ada-initialize-runtime-library (cross-prefix)
+  "Initializes the variables for the runtime library location.
+CROSS-PREFIX is the prefix to use for the gnatls command"
   (save-excursion
-    (set 'ada-xref-runtime-library-specs-path '())
-    (set 'ada-xref-runtime-library-ali-path '())
+    (setq ada-xref-runtime-library-specs-path '()
+	  ada-xref-runtime-library-ali-path   '())
     (set-buffer (get-buffer-create "*gnatls*"))
     (widen)
     (erase-buffer)
@@ -187,7 +206,8 @@ As always, the values of the project file are defined through properties.")
 	;;  Even if we get an error, delete the *gnatls* buffer
 	(unwind-protect
 	    (progn
-	      (call-process "gnatls" nil t nil "-v")
+	      (call-process (concat cross-prefix "gnatls")
+			    nil t nil "-v")
 	      (goto-char (point-min))
 
 	      ;;  Source path
@@ -233,24 +253,30 @@ edited file, minus extension but with directory, and ${full_current} is
 replaced by the name including the extension."
 
   (while (string-match "\\(-[^-\$IO]*[IO]\\)?\${\\([^}]+\\)}" cmd-string)
-    (let (value)
+    (let (value
+	  (name (match-string 2 cmd-string)))
       (cond
-       ((string= (match-string 2 cmd-string) "current")
-	(set 'value (file-name-sans-extension (buffer-file-name))))
-       ((string= (match-string 2 cmd-string) "full_current")
-	(set 'value (buffer-file-name)))
+       ((string= name "current")
+	(setq value (file-name-sans-extension (buffer-file-name))))
+       ((string= name "full_current")
+	(setq value (buffer-file-name)))
        (t
 	(save-match-data
-	  (set 'value (ada-xref-get-project-field
-		       (intern (match-string 2 cmd-string)))))))
+	  (setq value (ada-xref-get-project-field (intern name))))))
+
+      ;; Check if there is an environment variable with the same name
+      (if (null value)
+	  (if (not (setq value (getenv name)))
+	      (message (concat "No environment variable " name " found"))))
+		
       (cond
        ((null value)
-	(set 'cmd-string (replace-match "" t t cmd-string)))
+	(setq cmd-string (replace-match "" t t cmd-string)))
        ((stringp value)
-	(set 'cmd-string (replace-match value t t cmd-string)))
+	(setq cmd-string (replace-match value t t cmd-string)))
        ((listp value)
 	(let ((prefix (match-string 1 cmd-string)))
-	  (set 'cmd-string (replace-match
+	  (setq cmd-string (replace-match
 			    (mapconcat (lambda(x) (concat prefix x)) value " ")
 			    t t cmd-string)))))
       ))
@@ -268,17 +294,17 @@ replaced by the name including the extension."
 	   ;;  Try hard to find a default value for filename, so that the user
 	   ;;  can edit his project file even if the current buffer is not an
 	   ;;  Ada file or not even associated with a file
-	   (list 'filename        (expand-file-name
-				   (cond
-				    (file
-				     (ada-prj-get-prj-dir file))
-				    (ada-prj-default-project-file
-				     ada-prj-default-project-file)
-				    (t
-				     (message (concat "Not editing an Ada file,"
- 						    "and no default project "
- 						    "file specified!"))
-				     "")))
+	   (list 'filename (expand-file-name
+			    (cond
+			     (file
+			      (ada-prj-get-prj-dir file))
+			     (ada-prj-default-project-file
+			      ada-prj-default-project-file)
+			     (t
+			      (message (concat "Not editing an Ada file,"
+					       "and no default project "
+					       "file specified!"))
+			      "")))
 		 'build_dir       (file-name-as-directory (expand-file-name "."))
 		 'src_dir         (list ".")
 		 'obj_dir         (list ".")
@@ -289,8 +315,10 @@ replaced by the name including the extension."
 		 'bind_opt        ada-prj-default-bind-opt
 		 'link_opt        ada-prj-default-link-opt
 		 'gnatmake_opt    ada-prj-default-gnatmake-opt
+		 'gnatfind_opt    ada-prj-gnatfind-switches
 		 'main            (if file
-				      (file-name-sans-extension file)
+				      (file-name-nondirectory
+				       (file-name-sans-extension file))
 				    "")
 		 'main_unit       (if file
 				      (file-name-nondirectory
@@ -329,7 +357,7 @@ addition return the default paths."
 	file value)
 
     ;;  Get the project file (either the current one, or a default one)
-    (set 'file (or (assoc file-name ada-xref-project-files)
+    (setq file (or (assoc file-name ada-xref-project-files)
 		   (assoc nil ada-xref-project-files)))
 	
     ;;  If the file was not found, use the default values
@@ -428,13 +456,10 @@ All the directories are returned as absolute directories."
      ;; one (in case ada-xref-project-files is nil)
      (or ada-xref-project-files '(nil)))
 
-    (if (lookup-key ada-mode-map [menu-bar Ada Project])
-	(setcdr (lookup-key ada-mode-map [menu-bar Ada Project]) submenu)
-
-      ;;  Menu doesn't exist => create it (only for XEmacs)
-      (if (boundp 'running-xemacs)
-	  (funcall (symbol-function 'add-submenu)
-		   '("Ada") (list "Project" submenu) "Goto")))
+     (if (not ada-xemacs)
+         (if (lookup-key ada-mode-map [menu-bar Ada Project])
+             (setcdr (lookup-key ada-mode-map [menu-bar Ada Project])
+		     submenu)))
     ))
 
 
@@ -457,7 +482,8 @@ defined in the project file."
 	(dirs (ada-xref-get-src-dir-field)))
 
     (while dirs
-      (set 'list (append list (file-name-all-completions string (car dirs))))
+      (if (file-directory-p (car dirs))
+	  (set 'list (append list (file-name-all-completions string (car dirs)))))
       (set 'dirs (cdr dirs)))
     (cond ((equal flag 'lambda)
 	   (assoc string list))
@@ -591,6 +617,13 @@ name as was passed to `ada-create-menu'."
 		  (setq ada-xref-confirm-compile
 			(not ada-xref-confirm-compile))
 		  :style toggle :selected ada-xref-confirm-compile])
+	(if (string-match "gvd" ada-prj-default-debugger)
+	    (funcall (symbol-function 'add-menu-button)
+		     options-menu
+		     ["Tight Integration With Gnu Visual Debugger"
+		      (setq ada-tight-gvd-integration
+			    (not ada-tight-gvd-integration))
+		      :style toggle :selected ada-tight-gvd-integration]))
 	)
     
     ;; for Emacs
@@ -658,12 +691,17 @@ name as was passed to `ada-create-menu'."
 		     (setq ada-xref-other-buffer (not ada-xref-other-buffer)))
 		   :button (:toggle . ada-xref-other-buffer)) t)
 
-      ;;  Add an item to the standard Files menu
-      (define-key ada-mode-map [menu-bar files]
-	(lookup-key global-map [menu-bar files]))
-    
-      (define-key ada-mode-map [menu-bar files open-file-from-src-path]
-	'("Open File from Path..." . ada-find-file))
+      (if (string-match "gvd" ada-prj-default-debugger)
+	  (define-key-after options-menu [tightgvd]
+	    '(menu-item "Tight Integration With Gnu Visual Debugger"
+			(lambda()(interactive)
+			  (setq ada-tight-gvd-integration
+				(not ada-tight-gvd-integration)))
+			:button (:toggle . ada-tight-gvd-integration)) t))
+
+      (define-key ada-mode-map [menu-bar Ada Edit rem3] '("------------" . nil))
+      (define-key ada-mode-map [menu-bar Ada Edit open-file-from-src-path]
+	'("Search File on source path..." . ada-find-file))
       )
     )
   (ada-xref-update-project-menu)
@@ -802,7 +840,7 @@ The current buffer should be the ada-file buffer."
       (let (project src_dir obj_dir make_cmd comp_cmd check_cmd casing
 		    run_cmd debug_pre_cmd debug_post_cmd
             (ada-buffer (current-buffer)))
-	(set 'prj-file (expand-file-name prj-file))
+	(setq prj-file (expand-file-name prj-file))
 
 	;;  Initialize the project with the default values
 	(ada-xref-set-default-prj-values 'project (current-buffer))
@@ -864,18 +902,16 @@ The current buffer should be the ada-file buffer."
 					       (reverse check_cmd))))
 	(if run_cmd (set 'project (plist-put project 'run_cmd
 					     (reverse run_cmd))))
-	(if debug_post_cmd
-	    (set 'project (plist-put project 'debug_post_cmd
-				     (reverse debug_post_cmd))))
-	(if debug_pre_cmd
-	    (set 'project (plist-put project 'debug_pre_cmd
-				     (reverse debug_pre_cmd))))
+	(set 'project (plist-put project 'debug_post_cmd
+				 (reverse debug_post_cmd)))
+	(set 'project (plist-put project 'debug_pre_cmd
+				 (reverse debug_pre_cmd)))
 
 	;;  Delete the default project file from the list, if it is there.
 	;;  Note that in that case, this default project is the only one in
 	;;  the list
 	(if (assoc nil ada-xref-project-files)
-	    (set 'ada-xref-project-files nil))
+	    (setq ada-xref-project-files nil))
 	
 	;;  Memorize the newly read project file
 	(if (assoc prj-file ada-xref-project-files)
@@ -883,7 +919,7 @@ The current buffer should be the ada-file buffer."
 	  (add-to-list 'ada-xref-project-files (cons prj-file project)))
 
 	;;  Set the project file as the active one.
-	(set 'ada-prj-default-project-file prj-file)
+	(setq ada-prj-default-project-file prj-file)
 	
 	;; Sets up the compilation-search-path so that Emacs is able to
 	;; go to the source of the errors in a compilation buffer
@@ -897,7 +933,7 @@ The current buffer should be the ada-file buffer."
 	
 	;; Add the directories to the search path for ff-find-other-file
 	;; Do not add the '/' or '\' at the end
-	(set 'ada-search-directories
+	(setq ada-search-directories
 	     (append (mapcar 'directory-file-name compilation-search-path)
 		     ada-search-directories))
 	
@@ -954,9 +990,12 @@ ENTITY was first found the location given by FILE, LINE and COLUMN."
   ;;  processed (gnatfind \"+\":...).
   (let* ((quote-entity
 	  (if (= (aref entity 0) ?\")
-	      (concat "\\\"" (substring entity 1 -1) "\\\"")
+	      (if is-windows
+		  (concat "\\\"" (substring entity 1 -1) "\\\"")
+		(concat "'\"" (substring entity 1 -1) "\"'"))
 	    entity))
-	 (command (concat "gnatfind " ada-prj-gnatfind-switches " "
+	 (switches (ada-xref-get-project-field 'gnatfind_opt))
+	 (command (concat "gnatfind " switches " "
 			  quote-entity
                           (if file (concat ":" (file-name-nondirectory file)))
                           (if line (concat ":" line))
@@ -1070,7 +1109,18 @@ If OTHER-FRAME is non-nil, display the cross-reference in another frame."
   (let ((identlist (ada-read-identifier pos)))
     (condition-case nil
 	(ada-find-in-ali identlist other-frame)
-      (error (ada-find-in-src-path identlist other-frame)))))
+      (error
+       (let ((ali-file (ada-get-ali-file-name (ada-file-of identlist))))
+
+	 ;; If the ALI file was up-to-date, then we probably have a predefined
+	 ;; entity, whose references are not given by GNAT
+	 (if (and (file-exists-p ali-file)
+		  (file-newer-than-file-p ali-file (ada-file-of identlist)))
+	     (message "No cross-reference found. It might be a predefined entity.")
+
+	   ;; Else, look in every ALI file, except if the user doesn't want that
+	   (if ada-xref-search-with-egrep
+	       (ada-find-in-src-path identlist other-frame))))))))
 
 (defun ada-goto-declaration-other-frame (pos &optional other-frame)
   "Display the declaration of the identifier around POS.
@@ -1088,18 +1138,11 @@ The declation is shown in another frame if `ada-xref-other-buffer' is non-nil."
 	      machine
 	      command))))
 
-(defun ada-get-absolute-dir (dir root-dir)
-  "Returns the absolute directory corresponding to DIR.
-If DIR is a relative directory, the value of ROOT-DIR is added in front."
-  (if (file-name-absolute-p dir)
-      dir
-    (concat root-dir dir)))
-
 (defun ada-get-absolute-dir-list (dir-list root-dir)
   "Returns the list of absolute directories found in dir-list.
 If a directory is a relative directory, the value of ROOT-DIR is added in
 front."
-  (mapcar (lambda (x) (ada-get-absolute-dir x root-dir)) dir-list))
+  (mapcar (lambda (x) (expand-file-name x root-dir)) dir-list))
 
 (defun ada-set-environment ()
   "Return the new value for process-environment.
@@ -1114,13 +1157,13 @@ project file."
 	(set 'objects (concat path-separator objects)))
     (cons
      (concat "ADA_INCLUDE_PATH="
-	     (mapconcat (lambda(x) (ada-get-absolute-dir x build-dir))
+	     (mapconcat (lambda(x) (expand-file-name x build-dir))
 			(ada-xref-get-project-field 'src_dir)
 			path-separator)
 	     include)
      (cons
       (concat "ADA_OBJECTS_PATH="
-	      (mapconcat (lambda(x) (ada-get-absolute-dir x build-dir))
+	      (mapconcat (lambda(x) (expand-file-name x build-dir))
 			 (ada-xref-get-project-field 'obj_dir)
 			 path-separator)
 	      objects)
@@ -1135,7 +1178,7 @@ If ARG is not nil, ask for user confirmation."
 	(process-environment (ada-set-environment))
 	(compilation-scroll-output t))
 
-    (set 'compilation-search-path (ada-xref-get-src-dir-field))
+    (setq compilation-search-path (ada-xref-get-src-dir-field))
 
     ;;  If no project file was found, ask the user
     (unless cmd
@@ -1149,9 +1192,12 @@ If ARG is not nil, ask for user confirmation."
 	(setq cmd (read-from-minibuffer "enter command to compile: " cmd)))
 
     ;;  Insert newlines so as to separate the name of the commands to run
-    ;;  and the output of the commands.
-    (compile (concat cmd "\n\n"))
-  ))
+    ;;  and the output of the commands. this doesn't work with cmdproxy.exe,
+    ;;  which gets confused by newline characters.
+    (if (not (string-match "cmdproxy.exe" shell-file-name))
+	(setq cmd (concat cmd "\n\n")))
+    
+    (compile (ada-quote-cmd cmd))))
 
 (defun ada-compile-current (&optional arg prj-field)
   "Recompile the current file.
@@ -1165,7 +1211,7 @@ command, and should be either comp_cmd (default) or check_cmd."
 	 (process-environment (ada-set-environment))
 	 (compilation-scroll-output t))
     
-    (set 'compilation-search-path (ada-xref-get-src-dir-field))
+    (setq compilation-search-path (ada-xref-get-src-dir-field))
 
     (unless cmd
       (setq cmd '("") arg t))
@@ -1178,7 +1224,13 @@ command, and should be either comp_cmd (default) or check_cmd."
     (if (or ada-xref-confirm-compile arg)
 	(setq cmd (read-from-minibuffer "enter command to compile: " cmd)))
 
-    (compile (concat cmd "\n\n"))))
+    ;;  Insert newlines so as to separate the name of the commands to run
+    ;;  and the output of the commands. this doesn't work with cmdproxy.exe,
+    ;;  which gets confused by newline characters.
+    (if (not (string-match "cmdproxy.exe" shell-file-name))
+	(setq cmd (concat cmd "\n\n")))
+    
+    (compile (ada-quote-cmd cmd))))
 
 (defun ada-check-current (&optional arg)
   "Recompile the current file.
@@ -1203,12 +1255,12 @@ if ARG is not-nil, asks for user confirmation."
         (set 'command (list (file-name-sans-extension (buffer-name)))))
 
     ;; Modify the command to run remotely
-    (set 'command (ada-remote (mapconcat 'identity command
+    (setq command (ada-remote (mapconcat 'identity command
 					 ada-command-separator)))
     
     ;; Ask for the arguments to the command if required
     (if (or ada-xref-confirm-compile arg)
-	(set 'command (read-from-minibuffer "Enter command to execute: "
+	(setq command (read-from-minibuffer "Enter command to execute: "
 					    command)))
 
     ;; Run the command
@@ -1233,15 +1285,18 @@ if ARG is not-nil, asks for user confirmation."
     (switch-to-buffer "*run*")
     ))
 
-(defun ada-gdb-application (&optional arg)
+(defun ada-gdb-application (&optional arg executable-name)
   "Start the debugger on the application.
+EXECUTABLE-NAME, if non-nil, is debugged instead of the file specified in the
+project file.
 If ARG is non-nil, ask the user to confirm the command."
   (interactive "P")
   (let ((buffer (current-buffer))
-        gdb-buffer
 	cmd pre-cmd post-cmd)
     (ada-require-project-file)
-    (setq cmd      (ada-xref-get-project-field 'debug_cmd)
+    (setq cmd   (if executable-name
+		    (concat ada-prj-default-debugger " " executable-name)
+		  (ada-xref-get-project-field 'debug_cmd))
 	  pre-cmd  (ada-xref-get-project-field 'debug_pre_cmd)
 	  post-cmd (ada-xref-get-project-field 'debug_post_cmd))
 
@@ -1249,7 +1304,24 @@ If ARG is non-nil, ask the user to confirm the command."
     (if (not cmd)
 	(set 'cmd (concat ada-prj-default-debugger
 			  " "
-			  (file-name-sans-extension (buffer-file-name)))))
+			  (or executable-name
+			      (file-name-sans-extension (buffer-file-name))))))
+
+    ;;  For gvd, add an extra switch so that the Emacs window is completly
+    ;;  swallowed inside the Gvd one
+    (if (and ada-tight-gvd-integration
+	     (string-match "^[^ \t]*gvd" cmd))
+	;;  Start a new frame, so that when gvd exists we do not kill Emacs
+	;;  We make sure that gvd swallows the new frame, not the one the
+	;;  user has been using until now
+	;;  The frame is made invisible initially, so that GtkPlug gets a
+	;;  chance to fully manage it. Then it works fine with Enlightenment
+	;;  as well
+	(let ((frame (make-frame '((visibility . nil)))))
+	  (set 'cmd (concat
+		     cmd " --editor-window="
+		     (cdr (assoc 'outer-window-id (frame-parameters frame)))))
+	  (select-frame frame)))
 
     ;;  Add a -fullname switch
     ;;  Use the remote machine
@@ -1259,10 +1331,6 @@ If ARG is non-nil, ask the user to confirm the command."
     (if (or arg ada-xref-confirm-compile)
 	(set 'cmd (read-from-minibuffer "enter command to debug: " cmd)))
 
-    ;;  Set the variable gud-last-last-frame so that glide-debug can find
-    ;;  the name of the Ada file, and thus of the project file if needed.
-    (set 'gud-last-last-frame (cons ada-prj-default-project-file 1))
-
     (let (comint-exec
 	  in-post-mode
 	  gud-gdb-massage-args)
@@ -1270,7 +1338,10 @@ If ARG is non-nil, ask the user to confirm the command."
       ;;  Do not add -fullname, since we can have a 'rsh' command in front.
       (fset 'gud-gdb-massage-args (lambda (file args) args))
 
-      (set 'pre-cmd  (mapconcat 'identity pre-cmd  "\n"))
+      (set 'pre-cmd  (mapconcat 'identity pre-cmd  ada-command-separator))
+      (if (not (equal pre-cmd ""))
+	  (setq pre-cmd (concat pre-cmd ada-command-separator)))
+
       (set 'post-cmd (mapconcat 'identity post-cmd "\n"))
       (if post-cmd
 	(set 'post-cmd (concat post-cmd "\n")))
@@ -1283,10 +1354,17 @@ If ARG is non-nil, ask the user to confirm the command."
 		 (save-excursion
 		   (set 'compilation-buffer-name-function
 			(lambda(x) (buffer-name buffer)))
-		   (compile (concat ,pre-cmd "\n"
-				    command " "
-				    (mapconcat 'identity switches " ")))))
+		   (compile (ada-quote-cmd
+			     (concat ,pre-cmd
+				     command " "
+				     (mapconcat 'identity switches " "))))))
 	       ))
+
+      ;;  Tight integration should force the tty mode
+      (if (and (string-match "gvd" (comint-arguments cmd 0 0))
+	       ada-tight-gvd-integration
+	       (not (string-match "--tty" cmd)))
+	  (setq cmd (concat cmd "--tty")))
       
       (if (and (string-match "jdb" (comint-arguments cmd 0 0))
 	       (boundp 'jdb))
@@ -1295,64 +1373,78 @@ If ARG is non-nil, ask the user to confirm the command."
 
       ;;  Send post-commands to the debugger
       (process-send-string (get-buffer-process (current-buffer)) post-cmd)
-      )
-    
-    (set 'gdb-buffer (symbol-value 'gud-comint-buffer))
-    
-    ;;  Switch back to the source buffer
-    ;;  and Activate the debug part in the contextual menu
-    (switch-to-buffer buffer)
 
-    (if (functionp 'gud-make-debug-menu)
-	(funcall (symbol-function 'gud-make-debug-menu)))
+      ;;  Move to the end of the debugger buffer, so that it is automatically
+      ;;  scrolled from then on.
+      (end-of-buffer)
 
-    ;;  Warning: on Emacs >= 20.3.8, same-window-regexps includes gud-*,
-    ;;  so the following call to display buffer will select the
-    ;;  buffer instead of displaying it in another window
-    ;;  This is why the second argument to display-buffer is 't'
-    (display-buffer gdb-buffer t)
-    ))
+      ;;  Display both the source window and the debugger window (the former
+      ;;  above the latter). No need to show the debugger window unless it
+      ;;  is going to have some relevant information.
+      (if (or (not (string-match "gvd" (comint-arguments cmd 0 0)))
+	      (string-match "--tty" cmd))
+	  (split-window-vertically))
+      (switch-to-buffer buffer)
+      )))
 
 
 (defun ada-reread-prj-file (&optional filename)
-  "Force Emacs to parse a project file.
-Either FILENAME or the project file associated with the current buffer is
-parsed.
-This file is only read once, and never read again.
+  "Forces Emacs to read either FILENAME or the project file associated
+with the current buffer.
+Otherwise, this file is only read once, and never read again.
 Since the information in the project file is shared between all buffers, this
 automatically modifies the setup for all the Ada buffer that use this file."
   (interactive "P")
   (if filename
       (ada-parse-prj-file filename)
     (ada-parse-prj-file (ada-prj-find-prj-file)))
+
+  ;; Reread the location of the standard runtime library
+  (ada-initialize-runtime-library
+   (or (ada-xref-get-project-field 'cross-prefix) ""))
   )
 
 ;; ------ Private routines
 
 (defun ada-xref-current (file &optional ali-file-name)
   "Update the cross-references for FILE.
-This in fact recompiles FILE to create ALI-FILE-NAME."
+This in fact recompiles FILE to create ALI-FILE-NAME.
+This function returns the name of the file that was recompiled to generate
+the cross-reference information. Note that the ali file can then be deduced by
+replacing the file extension with .ali"
   ;; kill old buffer
   (if (and ali-file-name
            (get-file-buffer ali-file-name))
       (kill-buffer (get-file-buffer ali-file-name)))
-  ;; read the project file
-  (ada-require-project-file)
-  (let* ((cmd (ada-xref-get-project-field 'comp_cmd))
-	 (process-environment (ada-set-environment))
-	 (compilation-scroll-output t)
-	 (name      (ada-convert-file-name (buffer-file-name)))
-	 (body-name (ada-get-body-name name)))
+  
+  (let* ((name      (ada-convert-file-name file))
+	 (body-name (or (ada-get-body-name name) name)))
 
+    ;; Always recompile the body when we can. We thus temporarily switch to a
+    ;; buffer than contains the body of the unit
+    (save-excursion
+      (let ((body-visible (find-buffer-visiting body-name))
+	    process)
+	(if body-visible
+	    (set-buffer body-visible)
+	  (find-file body-name))
 
-    ;;  ???We should add -gnatQ here, and use ada-compile-file instead
-    
+	;; Execute the compilation. Note that we must wait for the end of the
+	;; process, or the ALI file would still not be available.
+	;; Unfortunately, the underlying `compile' command that we use is
+	;; asynchronous.
+	(ada-compile-current)
+	(setq process (get-buffer-process "*compilation*"))
 
-    ;; Always recompile the body when we can
-    (set 'body-name (or body-name name))
+	(while (and process
+		    (not (equal (process-status process) 'exit)))
+	  (sit-for 1))
 
-    ;; prompt for command to execute
-    (ada-compile-current)))
+	;; remove the buffer for the body if it wasn't there before
+	(unless body-visible
+	  (kill-buffer (find-buffer-visiting body-name)))
+	))
+    body-name))
 
 (defun ada-find-file-in-dir (file dir-list)
   "Search for FILE in DIR-LIST."
@@ -1466,21 +1558,26 @@ the project file."
       
       ;; If still not found, try to recompile the file
       (if (not ali-file-name)
-	  (progn
-	    ;; recompile only if the user asked for this
-	    (if ada-xref-create-ali
-		(ada-xref-current file ali-file-name))
-	    (error "Ali file not found. Recompile your file")))
+	  ;; recompile only if the user asked for this. and search the ali
+	  ;; filename again. We avoid a possible infinite recursion by
+	  ;; temporarily disabling the automatic compilation.
+	  
+	  (if ada-xref-create-ali
+	      (setq ali-file-name
+		    (concat (file-name-sans-extension (ada-xref-current file))
+			    ".ali"))
+
+	    (error "Ali file not found. Recompile your file"))
       
-    
-      ;; same if the .ali file is too old and we must recompile it
-      (if (and (file-newer-than-file-p file ali-file-name)
-	       ada-xref-create-ali)
-	  (ada-xref-current file ali-file-name))
-      
-      ;; else returns the correct absolute file name
+	
+	;; same if the .ali file is too old and we must recompile it
+	(if (and (file-newer-than-file-p file ali-file-name)
+		 ada-xref-create-ali)
+	    (ada-xref-current file ali-file-name)))
+
+      ;;  Always return the correct absolute file name
       (expand-file-name ali-file-name))
-    ))
+      ))
 
 (defun ada-get-ada-file-name (file original-file)
   "Create the complete file name (+directory) for FILE.
@@ -1626,11 +1723,13 @@ from the ali file (definition file and places where it is referenced)."
           ;; if we did not find it, it may be because the first reference
           ;; is not required to have a 'unit_number|' item included.
           ;; Or maybe we are already on the declaration...
-          (unless (re-search-forward (concat "^\\([a-zA-Z0-9_.\"]+[ *]\\)*"
-					     (ada-line-of identlist)
-					     "[^0-9]"
-					     (ada-column-of identlist))
-				     nil t)
+          (unless (re-search-forward
+		   (concat
+		    "^\\(\\([a-zA-Z0-9_.]+\\|\"[<>=+*-/a-z]\"\\)[ *]\\)*"
+		    (ada-line-of identlist)
+		    "[^0-9]"
+		    (ada-column-of identlist))
+		   nil t)
 	    
 	    ;; If still not found, then either the declaration is unknown
 	    ;; or the source file has been modified since the ali file was
@@ -2203,16 +2302,17 @@ This function typically is to be hooked into `ff-file-created-hooks'."
       )))
 
 (defun ada-xref-initialize ()
-  "Function called by ada-mode-hook to initialize the ada-xref.el package.
-For instance, it creates the gnat-specific menus, set some hooks for
+  "Function called by `ada-mode-hook' to initialize the ada-xref.el package.
+For instance, it creates the gnat-specific menus, sets some hooks for
 find-file...."
   (make-local-hook 'ff-file-created-hooks)
+  ;; This should really be an `add-hook'.  -stef
   (setq ff-file-created-hooks 'ada-make-body-gnatstub)
 
   ;; Read the project file and update the search path
   ;; before looking for the other file
   (make-local-hook 'ff-pre-find-hooks)
-  (add-hook 'ff-pre-find-hooks 'ada-require-project-file)
+  (add-hook 'ff-pre-find-hooks 'ada-require-project-file nil t)
 
   ;; Completion for file names in the mini buffer should ignore .ali files
   (add-to-list 'completion-ignored-extensions ".ali")
@@ -2221,6 +2321,19 @@ find-file...."
 
 ;; ----- Add to ada-mode-hook ---------------------------------------------
 
+;;  Use gvd or ddd as the default debugger if it was found
+;;  On windows, do not use the --tty switch for GVD, since this is
+;;  not supported. Actually, we do not use this on Unix either, since otherwise
+;;  there is no console window left in GVD, and people have to use the
+;;  Emacs one.
+;;  This must be done before initializing the Ada menu.
+(if (ada-find-file-in-dir "gvd" exec-path)
+    (set 'ada-prj-default-debugger "gvd ")
+  (if (ada-find-file-in-dir "gvd.exe" exec-path)
+     (set 'ada-prj-default-debugger "gvd ")
+  (if (ada-find-file-in-dir "ddd" exec-path)
+      (set 'ada-prj-default-debugger "ddd --tty -fullname -toolbar"))))
+
 ;;  Set the keymap once and for all, so that the keys set by the user in his
 ;;  config file are not overwritten every time we open a new file.
 (ada-add-ada-menu)
@@ -2228,14 +2341,8 @@ find-file...."
 
 (add-hook 'ada-mode-hook 'ada-xref-initialize)
 
-;;  Use gvd or ddd as the default debugger if it was found
-(if (ada-find-file-in-dir "gvd" exec-path)
-    (set 'ada-prj-default-debugger "gvd --tty")
-  (if (ada-find-file-in-dir "ddd" exec-path)
-      (set 'ada-prj-default-debugger "ddd --tty -fullname -toolbar")))
-
 ;;  Initializes the cross references to the runtime library
-(ada-initialize-runtime-library)
+(ada-initialize-runtime-library "")
 
 ;;  Add these standard directories to the search path
 (set 'ada-search-directories
@@ -2253,4 +2360,3 @@ find-file...."
 (provide 'ada-xref)
 
 ;;; ada-xref.el ends here
-
